// Hippo subprocess handler implementation.

#include "hippo.h"

extern "C" {
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
}


// The PID of the running process.
static pid_t subprocess_pid = 0;

// The playlist which needs notifying when a process dies.
static playlist *subprocess_playlist = NULL;


// Start a new process with the supplied command; the supplied playlist
// will get signalled when this song finishes.
void subprocess_start(const string& command, playlist *newlist) {
	if (subprocess_pid) subprocess_stop();

	pid_t pid = fork();
	if (!pid) {
		// This is now the child process.

		// This is really ugly, but much more reliable and
		// flexible than attempting to do parsing ourselves.
		char *s[3];
		s[0] = "/bin/sh";
		s[1] = "-c";
		s[2] = (char *)("exec " + command).c_str();
		execvp(s[0], s);

		cerr << "unable to start player: " << command;
		exit(0);
	}

	subprocess_pid = pid;
	subprocess_playlist = newlist;
}

// Stop the current process.
void subprocess_stop() {
	// Clear the pointer first so that the playlist doesn't get signalled.
	subprocess_playlist = NULL;

	if (subprocess_pid) {
		kill(subprocess_pid, SIGINT);

		// Sleep until the SIGCHLD handler gets the signal.
		while (subprocess_pid) pause();		
	}
}

// Handle SIGCHLD.
void subprocess_sigchld(int sig) {
	// Catch the signal to avoid zombies.
	wait(NULL);

	// Set the PID variable to 0. Note that subprocess_stop watches this to
	// know when the signal has been received.
	subprocess_pid = 0;

	// Signal the playlist that the song finished.
	if (subprocess_playlist) subprocess_playlist->song_finished();
}

// Initialise subprocess handling.
void subprocess_init() {
	struct sigaction act;
	
	act.sa_handler = subprocess_sigchld;
	sigemptyset(&act.sa_mask);
	
	// SA_NOCLDSTOP prevents us from getting a signal when the child is stopped.
	// SA_RESTART restarts (rather than kills) pending IO operations, which is
	// necessary for the text interface to work.
	act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
	
	sigaction(SIGCHLD, &act, NULL);
}
